/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package dpll;

/**
 *
 * @author prashanth
 */


import java.util.*;

public class DPLL 
{
/**
     * Convention: 
     * Every clause in new line, no newline after last clause
     * box is *
     * OR is |
     * NOT is ~
     * 
     */
    public static HashMap<String,String> result=new HashMap<String,String>();
    ArrayList<String> sigma;
    
    public DPLL(ArrayList<String> sigma)
    {
        this.sigma=sigma;
    }
    
    public boolean doDpll() throws UnsatisfiableException
    {
        //System.out.print("Size="+sigma.size());
        if(sigma.size()==0)
        {
            //System.out.println("Returning true");
            return true; //satisfiable
        }
        
        for(int i=0;i<sigma.size();i++)  
            if(sigma.get(i).contains("*"))
                throw (new UnsatisfiableException());  //unsatisfiable
        
               
        if(!doUnitPropogation())        
                return doSplitting();                
        
        return true;
    }
    
    
    public void removeClausesContaining(String literal)
    {
        
        for(int i=0;i<sigma.size();i++)
        {
            String[] literals=sigma.get(i).split("[|]");
      
            for(int j=0;j<literals.length;j++)
            {
                if(literals[j].equals(literal))
                {
                   
                    sigma.set(i, "-1");
                    break;
                }
            }
        }
        
        ArrayList<String> s=new ArrayList<String>();
        for(int i=0;i<sigma.size();i++)
        {
            if(!sigma.get(i).equals("-1"))
                s.add(sigma.get(i));
        }
        sigma=s;
    }
    
    public void removeNegationOfLiteral(String literal) throws UnsatisfiableException
    {
        String negation=null;
        if(literal.startsWith("~"))
            negation=literal.substring(1);
        else
            negation="~"+literal;

        
        for(int i=0;i<sigma.size();i++)
        {
            
            String[] literals=sigma.get(i).split("[|]");
            
            String reducedClause="";
            for(int j=0;j<literals.length;j++)
            {
                if(!literals[j].equals(negation))     //remove negation of literal                    
                {
                    //System.out.print("adding "+literals[j]+"\t");
                    reducedClause+="|"+literals[j];
                }
            }
           
            if(reducedClause.length()==0)
            {
                reducedClause="*";          //box, l and ~l occured
                sigma.set(i, reducedClause);
                throw (new UnsatisfiableException());
                //return;
            }
            else
                reducedClause=reducedClause.substring(1); //remove first |
            sigma.set(i, reducedClause);
            
            //System.out.println("reduced clause:"+reducedClause);
            
        }
        
            
    }
    
    public void printSigma(String msg)
    {
        System.out.print(msg);
        for(int i=0;i<sigma.size();i++)
            System.out.print(sigma.get(i)+"\t");
        System.out.println();
    }
    
    public void print(String literal, String truthVal)
    {
        if(literal.startsWith("~"))
        {
            if(truthVal.equals("T"))
            {
                System.out.println(literal.substring(1)+":F");
                result.put(literal.substring(1), "F");
            }
            else
            {
                System.out.println(literal.substring(1)+":T");
                 result.put(literal.substring(1), "T");
            }
        }
        else
        {
            System.out.println(literal+":"+truthVal);
            result.put(literal,truthVal);
        }
    }
    
    public void simplify(String unitClause) throws UnsatisfiableException
    {
        printSigma("Before simplify:");
        removeClausesContaining(unitClause);        
        removeNegationOfLiteral(unitClause);
        printSigma("After simplify:");
        for(int i=0;i<sigma.size();i++)  
            if(sigma.get(i).contains("*"))
                throw (new UnsatisfiableException());  //unsatisfiable        
       
    }
    
    public boolean doUnitPropogation() throws UnsatisfiableException
    {
        //System.out.println("Unit prop");
        for(int i=0;i<sigma.size();i++)
            if(isUnitClause(sigma.get(i)))
            {
                //System.out.println(sigma.get(i)+":T");
                print(sigma.get(i), "T");
                simplify(sigma.get(i));
                ArrayList<String> s= new ArrayList<String>(sigma);
                
                return doDpll();
            }
        return false;
    }
    
    public boolean doSplitting() throws UnsatisfiableException
    {
        for(int i=0;i<sigma.size();i++)
        {
            int posn=0;
            String l=getLiteral(sigma.get(i), posn);
            posn++;
            System.out.println("Trying "+l+":T");
            ArrayList<String> s=new ArrayList<String>(sigma);
            s.add(l);
            DPLL d=new DPLL(s);
            try
            {
                if(d.doDpll())
                {
                    //System.out.println(l+":T");
                    return true;
                }
            }
            catch(UnsatisfiableException e)
            {
                System.out.println("Trial "+l+":T didn't work, Trying F");
                s=new ArrayList<String>(sigma);
                if(l.startsWith("~"))
                    s.add(l.substring(1));
                else
                    s.add("~"+l);
                
                d=new DPLL(s);
                
                if(d.doDpll())
                {
                    //System.out.println(l+":F");                           
                    return true;
                }
                
                return false;
            }
           
        }
        return false;
    }
    
    
    public String getLiteral(String clause, int posn)
    {
        String[] literals=clause.split("[|]");
        //System.out.println("Length:"+literals.length+"Literal,posn="+literals[posn]+posn);
        return literals[posn];
    }
    
    
   
    public boolean isUnitClause(String clause)
    {
        //Unit clause: L, ~L
        if(clause.length()<3)  
            return true;
        return false;           
            
    }

}
